/* -*-C-*-
*******************************************************************************
*
* File:         ComposeHelper.m
* RCS:          /usr/local/sources/CVS/EnhanceMail/ComposeHelper.m,v 1.5 1999/12/23 19:04:15 tom Exp
* Description:  
* Author:       Carl Edman
* Created:      Fri Oct 13 11:48:05 1995
* Modified:     Tue Jun 25 20:02:38 1996 (Carl Edman) cedman@capitalist.princeton.edu
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 1995, but otherwise this file is perfect freeware.
*
*******************************************************************************
*/

#import "EnhanceMail.h"
#import "ComposeHelper.h"
#import "Preferences.h"
#import "PGP.h"
#import "regexp.h"

@implementation EnhanceComposeHelper

+ finishLoading:(struct mach_header *)header
{
   [NXApp addComposeHelperClass:self];
   return self;
}


- init
{
   if (self=[super init])
   {
      if (EnhanceUsePGP)
      {
	 char path[MAXPATHLEN];

	 if ([EnhanceBundle getPath:path forResource:"EncryptComposeHelper" ofType:"nib"])
	    [NXApp loadNibFile:path owner:self];

	 offscreenWindow = [accessoryBox window];

	 [encryptSwitch setState:EnhancePGPEncrypt];
	 [signSwitch setState:EnhancePGPSign];
	 [useKeyButton setEnabled:[encryptSwitch state]];
	 [[useKeyText window] setFloatingPanel:YES];
      }
   }
   return self;
}

- free
{
   [[[useKeyText window] close] free];
   [offscreenWindow free];
   return [super free];
}


- toggleEncrypt:sender
{
   [useKeyButton setEnabled:[encryptSwitch state]];
   if ([encryptSwitch state] == NO) [[useKeyText window] orderOut:self];
   return self;
}

- showPGPKeyPanel:sender
{
   const char *title = [[sender window] title]; // ...in Compose window by now.

   if (title) [[useKeyText window] setTitle:title];

   [useKeyText selectText:sender];
   [[useKeyText window] makeKeyAndOrderFront:sender];
   return self;
}

- (void)y2kFixDateHeader:(MailMessage *)mes
{
   static regexp *daterx;
   const char *date = [mes headerValueForKey:"Date"];

   if (!date) return;

   if (!daterx)
      daterx = regcomp("^( *[^ ,]+, *[ 0-9][0-9] +[^ ]+ )([0-9][0-9]+)( +.*)");

   if (regexec(daterx, date))
   {
      int year = atoi(daterx->startp[2]);

      if (year < 200)
      {
	 char subst[20];
	 char *fixedDate = alloca(daterx->endp[0] - daterx->startp[0] + 2 + 1);

	 if (year < 70) year += 2000; else year += 1900;

	 sprintf(subst, "\\1%d\\3", year);
	 regsub(daterx, subst, fixedDate);
	 [mes setHeaderKey:"Date" value:fixedDate];
      }
   }
}

- (View *)composeView
{
   if (EnhanceUsePGP) return accessoryBox;
   return nil;
}

- (StringList *)recipientKeys
{
   StringList *result = nil;
   const char *keyValue = [useKeyText stringValue];

   if (keyValue)
   {
      while (*keyValue == ' ') keyValue++;
      if (*keyValue)
      {
	 result = [[StringList alloc] init];
	 [result addParsedAddressesFrom:keyValue];
	 if ([result count] == 0) result = [result free];
      }
   }
   return result;
}

- (BOOL)willDeliverMessage:(MailMessage *)mes to:(StringList *)rcpt
{
   BOOL result = YES;

   [self y2kFixDateHeader:mes];

   if (EnhanceUsePGP)
   {
      StringList *recipientKeys = nil;
      BOOL shouldSign = [signSwitch state];
      BOOL shouldEncrypt = [encryptSwitch state];

      if (shouldEncrypt)
      {
	 recipientKeys = [self recipientKeys];
	 if (recipientKeys) rcpt = recipientKeys;
      }
      result = [[EnhancePGP new] encodePGP:mes to:rcpt 
		  sign:shouldSign encrypt:shouldEncrypt];

      if (recipientKeys) [recipientKeys free];
   }
   return result;
}

@end // EnhanceComposeHelper
