/* -*-ObjC-*-
*******************************************************************************
*
* File:         MailBox.m
* RCS:          /usr/local/sources/CVS/EnhanceMail/MailBox.m,v 1.12 1999/11/12 23:24:47 tom Exp
* Description:  
* Author:       Carl Edman
* Created:      Fri Oct 13 11:48:05 1995
* Modified:     Sun May 30 21:02:03 1999  Tom Hageman
* Language:     ObjC
* Package:      N/A
* Status:       Experimental
*
* (C) Copyright 1995, but otherwise this file is perfect freeware.
*
*******************************************************************************
*/

#import "EnhanceMail.h"
#import "Preferences.h"
#import "MailBox.h"
#import "MailIndex.h"
#import "MailReader.h"
#import	"TransferPanel.h"
#import <libc.h>
#import <indexing/indexing.h>

#define MSG_OPEN_MAILBOX NXLoadLocalizedStringFromTableInBundle("Alerts", nil, "Open Mailbox", NULL)
#define MSG_REBUILD_INDEX NXLocalizedStringFromTableInBundle("Localizable", EnhanceBundle, "Rebuilding full-text index for %s (%u %s).  Please wait...", NULL, Message for full text index progressor - arguments: Mailbox name; number of messages; 'message' or 'messages')
#define MSG_UPDATE_INDEX NXLocalizedStringFromTableInBundle("Localizable", EnhanceBundle, "Updating full-text index for %s (%u %s).  Please wait...", NULL, Message for full text index progressor - arguments: Mailbox name; number of messages; 'message' or 'messages')

/* This is to keep genstrings from breaking. */
#define LOCALIZE(TABLE,BUNDLE,KEY,VALUE) NXLoadLocalized\
StringFromTableInBundle(TABLE,BUNDLE,KEY,VALUE)


static HashTable *StoreNameHash=nil;
static HashTable *ManagerHash=nil;

@implementation EnhanceMailBox

+ finishLoading:(struct mach_header *)header
{
   [self poseAs:[self superclass]];
   StoreNameHash=[[HashTable alloc] initKeyDesc:"@" valueDesc:"*"];
   ManagerHash=[[HashTable alloc] initKeyDesc:"@" valueDesc:"@"];
   return self;
}

+ startUnloading
{
   StoreNameHash=[StoreNameHash free];
   ManagerHash=[ManagerHash free];
   return self;
}

- initMailBox:(const char *)name
{
   char *storeName=0;
   id manager=nil;
   id ret=[super initMailBox:name];
   
   if (!dirname) return ret;
   storeName=strcpy(malloc(strlen(dirname)+20),dirname);
   strcat(storeName,"/content.store");
   [StoreNameHash insertKey:self value:storeName];

   manager=[[IXRecordManager alloc] initFromName:"content" inFile:storeName forWriting:YES];
   if (manager == nil) manager=[[IXRecordManager alloc] initFromName:"content" inFile:storeName forWriting:NO];

   [ManagerHash insertKey:self value:manager];

   [self maybeUpdateStore];

   [[TransferPanel new] updateBrowserForMailBoxNamed:dirname];
   return ret;
}

- free
{
   char *storeName=[StoreNameHash valueForKey:self];
   id manager=[ManagerHash valueForKey:self];

   [StoreNameHash removeKey:self];
   if (storeName) free(storeName);
   
   [ManagerHash removeKey:self];
   if (manager)
   {
      id store=[manager store];
    NX_DURING
      [store startTransaction];
      [store compact];
      [store commitTransaction];
    NX_HANDLER
      // XXX exception is probably due to corrupted index file.
      NXLogError("error while compacting index in %s"
		 " (exception code=%d data1=%s data2=%s)",
		 dirname,
		 NXLocalHandler.code, NXLocalHandler.data1, NXLocalHandler.data2);
    NX_ENDHANDLER
      manager=[manager free];
   }
   
   return [super free];
}

- (int)numberOfNonIndexedMessages
{
   id manager=[ManagerHash valueForKey:self];
   struct u_mailboxstatus stat;
   int count;

   if (manager==nil) return 0;
   
   [self getMailboxStatus:&stat];

 NX_DURING
   count=[manager count]-[manager attributeCount];
 NX_HANDLER
   // XXX exception is probably due to corrupted index file.
   NXLogError("error while determining indexed message count in %s"
	      " (exception code=%d data1=%s data2=%s)",
	      dirname,
	      NXLocalHandler.code, NXLocalHandler.data1, NXLocalHandler.data2);
   return 0;
 NX_ENDHANDLER

   if (stat.msgno<=count) return 0;
   
   return stat.msgno-count;
}

- createStore
{
   char *storeName=[StoreNameHash valueForKey:self];
   id manager=[ManagerHash valueForKey:self];
   id parser=nil;

   if (!storeName) return nil;
   if (manager==nil)
   {
    NX_DURING
      manager=[[IXRecordManager alloc] initWithName:"content" inFile:storeName];

      [[manager store] startTransaction];
      [manager addAttributeNamed:"msgId" forSelector:@selector(msgId)];
      [manager setTargetClass:[MailIndex class] forAttributeNamed:"msgId"];
      [manager addAttributeNamed:"Content" forSelector:@selector(messageContent)];
      [manager setTargetClass:[MailIndex class] forAttributeNamed:"Content"];
      parser=[[IXAttributeParser alloc] init];
      [manager setParser:parser forAttributeNamed:"Content"];
#if 0
      [manager addAttributeNamed:"Subject" forSelector:@selector(messageSubject)];
      [manager setTargetClass:[MailIndex class] forAttributeNamed:"Subject"];
      [manager addAttributeNamed:"From" forSelector:@selector(messageFrom)];
      [manager setTargetClass:[MailIndex class] forAttributeNamed:"From"];
      [manager addAttributeNamed:"To" forSelector:@selector(messageTo)];
      [manager setTargetClass:[MailIndex class] forAttributeNamed:"To"];
      [manager addAttributeNamed:"Cc" forSelector:@selector(messageCc)];
      [manager setTargetClass:[MailIndex class] forAttributeNamed:"Cc"];
      [manager addAttributeNamed:"Date" forSelector:@selector(messageDate)];
      [manager setTargetClass:[MailIndex class] forAttributeNamed:"Date"];
      [manager addAttributeNamed:"Message-Id" forSelector:@selector(messageId)];
      [manager setTargetClass:[MailIndex class] forAttributeNamed:"Message-Id"];
      [manager addAttributeNamed:"Next-Reference" forSelector:@selector(messageNextReference)];
      [manager setTargetClass:[MailIndex class] forAttributeNamed:"Next-Reference"];
#endif
      [[manager store] commitTransaction];
    NX_HANDLER
      NXLogError("error while creating index in %s"
		 " (exception code=%d data1=%s data2=%s)",
		 dirname,
		 NXLocalHandler.code, NXLocalHandler.data1, NXLocalHandler.data2);
      manager=[manager free];
      return nil;
    NX_ENDHANDLER
      [ManagerHash insertKey:self value:manager];
   }
   return [self maybeUpdateStore];   
}

- updateStore
{
   id manager=[ManagerHash valueForKey:self];
   struct u_mailboxstatus stat;
   int count;
   Progressor *progress;
   MailIndex *record;
   char buf[MAXPATHLEN+1];

   if (manager==nil) return nil;
   [self getMailboxStatus:&stat];
 NX_DURING
   count=[manager count]-[manager attributeCount];
 NX_HANDLER
   // XXX exception is probably due to corrupted index file.
   NXLogError("error while determining indexed message count in %s"
	      " (exception code=%d data1=%s data2=%s)",
	      dirname,
	      NXLocalHandler.code, NXLocalHandler.data1, NXLocalHandler.data2);
   return nil;
 NX_ENDHANDLER
   if (stat.msgno<=count) return self;
   [[manager store] startTransaction];
   
   sprintf(buf, (count > 0 ? MSG_UPDATE_INDEX : MSG_REBUILD_INDEX), dirname, (stat.msgno-count),
	   LOCALIZE("UI", nil, ((stat.msgno-count) == 1 ? "message" : "messages"), NULL));
   progress=[Progressor newWithTitle:MSG_OPEN_MAILBOX message:buf];
   [progress setMinValue:count];
   [progress setMaxValue:stat.msgno];
   [progress setProgress:count];
   [progress setShowProgress:YES];
   [progress beginModalSession];
   
   [MailIndex setMailBox:self];
   for(;(count<stat.msgno) && ![progress cancelled];count++)
   {
      record=[[MailIndex alloc] initMsgId:count];
      NX_DURING
      {
	 [manager addRecord:record];
      }
      NX_HANDLER
      {
	 // XXX exception is probably due to corrupted index file.
	 // XXX Show alert panel instead?
	 NXLogError("error while indexing message #%d in %s"
		    " (exception code=%d data1=%s data2=%s)",
		    count, dirname,
		    NXLocalHandler.code, NXLocalHandler.data1, NXLocalHandler.data2);
	 [progress cancel:self];
      }
      NX_ENDHANDLER
      record=[record free];
      [progress setProgress:count];
   }
   [MailIndex setMailBox:nil];
   
   [progress endModalSession];

 NX_DURING
   [[manager store] commitTransaction];
 NX_HANDLER
   // XXX exception is probably due to corrupted index file.
   NXLogError("error while indexing in %s"
	      " (exception code=%d data1=%s data2=%s)",
	      dirname,
	      NXLocalHandler.code, NXLocalHandler.data1, NXLocalHandler.data2);
 NX_ENDHANDLER

   return (count>=stat.msgno) ? self : nil;
}

- maybeUpdateStore
{
   if ([self numberOfNonIndexedMessages] > 25) return [self updateStore];
   // XXX hardcoded parameter.
   return nil;
}

- deleteStore
{
   char *storeName=[StoreNameHash valueForKey:self];
   id manager=[ManagerHash valueForKey:self];
      
   if (!storeName)
      return nil;

   if (manager)
      [ManagerHash insertKey:self value:(manager=[manager free])];

   if (unlink(storeName)==-1)
      return nil;

   return self;
}

- (BOOL)hasStore
{
   id manager=[ManagerHash valueForKey:self];
      
   return (manager==nil) ? NO : YES;
}

- (int)storeSize
{
   char *storeName=[StoreNameHash valueForKey:self];
   id manager=[ManagerHash valueForKey:self];
   struct stat buf;
   
   if ((manager==nil)||(storeName==0)||(stat(storeName,&buf)==-1)) return -1;
   
   return buf.st_size;
}

- (void)maybeUpdateToc
{
   [super maybeUpdateToc];
   [self maybeUpdateStore];
}

- repairTocFile:(int)start
{
   id ret=[super repairTocFile:start];
   [self maybeUpdateStore];
   return ret;
}

- query:(const char *)expr
{
   id manager,parser,query,list=nil;

   manager=[ManagerHash valueForKey:self];
   if (manager==nil) return nil;
   parser=[manager parserForAttributeNamed:"Content"];
   if (parser==nil) return nil;
   [self updateStore];
   query=[[IXAttributeQuery alloc] initQueryString:expr andAttributeParser:parser];
   if (query==nil) return nil;
 NX_DURING
   list=[query evaluateFor:manager];
 NX_HANDLER
   // XXX exception is probably due to corrupted index file.
   NXLogError("error while evaluating index query in %s"
	      " (exception code=%d data1=%s data2=%s)",
	      dirname,
	      NXLocalHandler.code, NXLocalHandler.data1, NXLocalHandler.data2);
 NX_ENDHANDLER
   [query free];
   return list;
}

#if 0
- flushMsgState;
{
   //puts("***flushMsgState");
   return [super flushMsgState];
}
#endif

#if 0 // Better handled in [MailReader updateStatus] ?
- (BOOL)setState:(char)fp16 forMsg:(int)fp20 flush:(BOOL)fp24;
{
   //puts("***setState");
   [[self mailReader] updateTransferPanel];
   return [super setState:fp16 forMsg:fp20 flush:fp24];
}
#endif

@end // EnhanceMailBox


@interface EnhanceIXStore : IXStore
/* Exceptions during IXStore -free escape from it (even when exceptions
   are disabled) , leading to all kinds of headache.  So we override -free
   to catch the exception, and pose as the real thing... */
- free;
@end


@implementation EnhanceIXStore : IXStore

+ finishLoading:(struct mach_header *)header
{
   [self poseAs:[self superclass]];
   return self;
}

- free
{
   NX_DURING
   {
      [super free];
   }
   NX_HANDLER
   {
      // XXX exception is probably due to corrupted index file.
      NXLogError("-[IXStore free]: Exception occurred, will retry after aborting pending transactions"
		 " (exception code=%d data1=%s data2=%s)",
		 NXLocalHandler.code, NXLocalHandler.data1, NXLocalHandler.data2);

      /* Retry, but abort pending transactions first. */
      NX_DURING
      {
	 while ([self nestingLevel] > 0) [self abortTransaction];
	 [super free];
      }
      NX_HANDLER
      {
	 // XXX exception is probably due to corrupted index file.
	 NXLogError("-[IXStore free]: Exception occurred, last-ditch attempt may leave memory leaks"
		    " (exception code=%d data1=%s data2=%s)",
		    NXLocalHandler.code, NXLocalHandler.data1, NXLocalHandler.data2);

	 /* Last-ditch attempt: invoke IXStore's super. */
	 (*[[[self superclass] superclass] instanceMethodFor:_cmd])(self, _cmd);
      }
      NX_ENDHANDLER
   }
   NX_ENDHANDLER
   return nil;
}

@end // EnhanceIXStore
